# DO NOT EDIT! This file is automatically generated
import marshmallow
import marshmallow_enum

from commercetools import helpers, types
from commercetools._schemas._common import (
    BaseResourceSchema,
    LocalizedStringField,
    ReferenceSchema,
    ResourceIdentifierSchema,
)
from commercetools._schemas._type import FieldContainerField

__all__ = [
    "CustomLineItemReturnItemSchema",
    "DeliveryItemSchema",
    "DeliverySchema",
    "DiscountedLineItemPriceDraftSchema",
    "ItemStateSchema",
    "LineItemImportDraftSchema",
    "LineItemReturnItemSchema",
    "OrderAddDeliveryActionSchema",
    "OrderAddItemShippingAddressActionSchema",
    "OrderAddParcelToDeliveryActionSchema",
    "OrderAddPaymentActionSchema",
    "OrderAddReturnInfoActionSchema",
    "OrderChangeOrderStateActionSchema",
    "OrderChangePaymentStateActionSchema",
    "OrderChangeShipmentStateActionSchema",
    "OrderFromCartDraftSchema",
    "OrderImportCustomLineItemStateActionSchema",
    "OrderImportDraftSchema",
    "OrderImportLineItemStateActionSchema",
    "OrderPagedQueryResponseSchema",
    "OrderReferenceSchema",
    "OrderRemoveDeliveryActionSchema",
    "OrderRemoveItemShippingAddressActionSchema",
    "OrderRemoveParcelFromDeliveryActionSchema",
    "OrderRemovePaymentActionSchema",
    "OrderResourceIdentifierSchema",
    "OrderSchema",
    "OrderSetBillingAddressActionSchema",
    "OrderSetCustomFieldActionSchema",
    "OrderSetCustomLineItemCustomFieldActionSchema",
    "OrderSetCustomLineItemCustomTypeActionSchema",
    "OrderSetCustomLineItemShippingDetailsActionSchema",
    "OrderSetCustomTypeActionSchema",
    "OrderSetCustomerEmailActionSchema",
    "OrderSetCustomerIdActionSchema",
    "OrderSetDeliveryAddressActionSchema",
    "OrderSetDeliveryItemsActionSchema",
    "OrderSetLineItemCustomFieldActionSchema",
    "OrderSetLineItemCustomTypeActionSchema",
    "OrderSetLineItemShippingDetailsActionSchema",
    "OrderSetLocaleActionSchema",
    "OrderSetOrderNumberActionSchema",
    "OrderSetParcelItemsActionSchema",
    "OrderSetParcelMeasurementsActionSchema",
    "OrderSetParcelTrackingDataActionSchema",
    "OrderSetReturnPaymentStateActionSchema",
    "OrderSetReturnShipmentStateActionSchema",
    "OrderSetShippingAddressActionSchema",
    "OrderSetStoreActionSchema",
    "OrderTransitionCustomLineItemStateActionSchema",
    "OrderTransitionLineItemStateActionSchema",
    "OrderTransitionStateActionSchema",
    "OrderUpdateActionSchema",
    "OrderUpdateItemShippingAddressActionSchema",
    "OrderUpdateSchema",
    "OrderUpdateSyncInfoActionSchema",
    "ParcelDraftSchema",
    "ParcelMeasurementsSchema",
    "ParcelSchema",
    "PaymentInfoSchema",
    "ProductVariantImportDraftSchema",
    "ReturnInfoSchema",
    "ReturnItemDraftSchema",
    "ReturnItemSchema",
    "ShippingInfoImportDraftSchema",
    "StagedOrderUpdateActionSchema",
    "SyncInfoSchema",
    "TaxedItemPriceDraftSchema",
    "TrackingDataSchema",
]


class DeliveryItemSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.DeliveryItem`."""

    id = marshmallow.fields.String(allow_none=True)
    quantity = marshmallow.fields.Integer(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.DeliveryItem(**data)


class DeliverySchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Delivery`."""

    id = marshmallow.fields.String(allow_none=True)
    created_at = marshmallow.fields.DateTime(allow_none=True, data_key="createdAt")
    items = helpers.LazyNestedField(
        nested="commercetools._schemas._order.DeliveryItemSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    parcels = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ParcelSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Delivery(**data)


class DiscountedLineItemPriceDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.DiscountedLineItemPriceDraft`."""

    value = helpers.LazyNestedField(
        nested="commercetools._schemas._common.MoneySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    included_discounts = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.DiscountedLineItemPortionSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        data_key="includedDiscounts",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.DiscountedLineItemPriceDraft(**data)


class ItemStateSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ItemState`."""

    quantity = marshmallow.fields.Integer(allow_none=True)
    state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ItemState(**data)


class LineItemImportDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.LineItemImportDraft`."""

    product_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="productId"
    )
    name = LocalizedStringField(allow_none=True)
    variant = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ProductVariantImportDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    price = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    quantity = marshmallow.fields.Integer(allow_none=True)
    state = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ItemStateSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    supply_channel = helpers.LazyNestedField(
        nested="commercetools._schemas._channel.ChannelResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="supplyChannel",
    )
    distribution_channel = helpers.LazyNestedField(
        nested="commercetools._schemas._channel.ChannelResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="distributionChannel",
    )
    tax_rate = helpers.LazyNestedField(
        nested="commercetools._schemas._tax_category.TaxRateSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="taxRate",
    )
    custom = helpers.LazyNestedField(
        nested="commercetools._schemas._type.CustomFieldsDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    shipping_details = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.ItemShippingDetailsDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingDetails",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.LineItemImportDraft(**data)


class OrderFromCartDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.OrderFromCartDraft`."""

    id = marshmallow.fields.String(allow_none=True)
    version = marshmallow.fields.Integer(allow_none=True)
    order_number = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="orderNumber"
    )
    payment_state = marshmallow_enum.EnumField(
        types.PaymentState, by_value=True, missing=None, data_key="paymentState"
    )
    shipment_state = marshmallow_enum.EnumField(
        types.ShipmentState, by_value=True, missing=None, data_key="shipmentState"
    )
    order_state = marshmallow_enum.EnumField(
        types.OrderState, by_value=True, missing=None, data_key="orderState"
    )
    state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.OrderFromCartDraft(**data)


class OrderImportDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.OrderImportDraft`."""

    order_number = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="orderNumber"
    )
    customer_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="customerId"
    )
    customer_email = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="customerEmail"
    )
    line_items = helpers.LazyNestedField(
        nested="commercetools._schemas._order.LineItemImportDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
        data_key="lineItems",
    )
    custom_line_items = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.CustomLineItemDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
        data_key="customLineItems",
    )
    total_price = helpers.LazyNestedField(
        nested="commercetools._schemas._common.MoneySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="totalPrice",
    )
    taxed_price = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.TaxedPriceDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="taxedPrice",
    )
    shipping_address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingAddress",
    )
    billing_address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="billingAddress",
    )
    customer_group = helpers.LazyNestedField(
        nested="commercetools._schemas._customer_group.CustomerGroupResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="customerGroup",
    )
    country = marshmallow.fields.String(allow_none=True, missing=None)
    order_state = marshmallow_enum.EnumField(
        types.OrderState, by_value=True, missing=None, data_key="orderState"
    )
    shipment_state = marshmallow_enum.EnumField(
        types.ShipmentState, by_value=True, missing=None, data_key="shipmentState"
    )
    payment_state = marshmallow_enum.EnumField(
        types.PaymentState, by_value=True, missing=None, data_key="paymentState"
    )
    shipping_info = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ShippingInfoImportDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingInfo",
    )
    completed_at = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="completedAt"
    )
    custom = helpers.LazyNestedField(
        nested="commercetools._schemas._type.CustomFieldsDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    inventory_mode = marshmallow_enum.EnumField(
        types.InventoryMode, by_value=True, missing=None, data_key="inventoryMode"
    )
    tax_rounding_mode = marshmallow_enum.EnumField(
        types.RoundingMode, by_value=True, missing=None, data_key="taxRoundingMode"
    )
    item_shipping_addresses = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
        data_key="itemShippingAddresses",
    )
    store = helpers.LazyNestedField(
        nested="commercetools._schemas._store.StoreResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    origin = marshmallow_enum.EnumField(types.CartOrigin, by_value=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.OrderImportDraft(**data)


class OrderPagedQueryResponseSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.OrderPagedQueryResponse`."""

    limit = marshmallow.fields.Integer(allow_none=True)
    count = marshmallow.fields.Integer(allow_none=True)
    total = marshmallow.fields.Integer(allow_none=True, missing=None)
    offset = marshmallow.fields.Integer(allow_none=True)
    results = helpers.LazyNestedField(
        nested="commercetools._schemas._order.OrderSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.OrderPagedQueryResponse(**data)


class OrderReferenceSchema(ReferenceSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderReference`."""

    obj = helpers.LazyNestedField(
        nested="commercetools._schemas._order.OrderSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type_id"]
        return types.OrderReference(**data)


class OrderResourceIdentifierSchema(ResourceIdentifierSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderResourceIdentifier`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type_id"]
        return types.OrderResourceIdentifier(**data)


class OrderSchema(BaseResourceSchema):
    """Marshmallow schema for :class:`commercetools.types.Order`."""

    id = marshmallow.fields.String(allow_none=True)
    version = marshmallow.fields.Integer(allow_none=True)
    created_at = marshmallow.fields.DateTime(allow_none=True, data_key="createdAt")
    last_modified_at = marshmallow.fields.DateTime(
        allow_none=True, data_key="lastModifiedAt"
    )
    last_modified_by = helpers.LazyNestedField(
        nested="commercetools._schemas._common.LastModifiedBySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="lastModifiedBy",
    )
    created_by = helpers.LazyNestedField(
        nested="commercetools._schemas._common.CreatedBySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="createdBy",
    )
    completed_at = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="completedAt"
    )
    order_number = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="orderNumber"
    )
    customer_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="customerId"
    )
    customer_email = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="customerEmail"
    )
    anonymous_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="anonymousId"
    )
    store = helpers.LazyNestedField(
        nested="commercetools._schemas._store.StoreKeyReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    line_items = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.LineItemSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        data_key="lineItems",
    )
    custom_line_items = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.CustomLineItemSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        data_key="customLineItems",
    )
    total_price = helpers.Discriminator(
        discriminator_field=("type", "type"),
        discriminator_schemas={
            "centPrecision": "commercetools._schemas._common.CentPrecisionMoneySchema",
            "highPrecision": "commercetools._schemas._common.HighPrecisionMoneySchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="totalPrice",
    )
    taxed_price = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.TaxedPriceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="taxedPrice",
    )
    shipping_address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingAddress",
    )
    billing_address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="billingAddress",
    )
    tax_mode = marshmallow_enum.EnumField(
        types.TaxMode, by_value=True, missing=None, data_key="taxMode"
    )
    tax_rounding_mode = marshmallow_enum.EnumField(
        types.RoundingMode, by_value=True, missing=None, data_key="taxRoundingMode"
    )
    customer_group = helpers.LazyNestedField(
        nested="commercetools._schemas._customer_group.CustomerGroupReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="customerGroup",
    )
    country = marshmallow.fields.String(allow_none=True, missing=None)
    order_state = marshmallow_enum.EnumField(
        types.OrderState, by_value=True, data_key="orderState"
    )
    state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    shipment_state = marshmallow_enum.EnumField(
        types.ShipmentState, by_value=True, missing=None, data_key="shipmentState"
    )
    payment_state = marshmallow_enum.EnumField(
        types.PaymentState, by_value=True, missing=None, data_key="paymentState"
    )
    shipping_info = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.ShippingInfoSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingInfo",
    )
    sync_info = helpers.LazyNestedField(
        nested="commercetools._schemas._order.SyncInfoSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        data_key="syncInfo",
    )
    return_info = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ReturnInfoSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
        data_key="returnInfo",
    )
    discount_codes = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.DiscountCodeInfoSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
        data_key="discountCodes",
    )
    last_message_sequence_number = marshmallow.fields.Integer(
        allow_none=True, data_key="lastMessageSequenceNumber"
    )
    cart = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.CartReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    custom = helpers.LazyNestedField(
        nested="commercetools._schemas._type.CustomFieldsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    payment_info = helpers.LazyNestedField(
        nested="commercetools._schemas._order.PaymentInfoSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="paymentInfo",
    )
    locale = marshmallow.fields.String(allow_none=True, missing=None)
    inventory_mode = marshmallow_enum.EnumField(
        types.InventoryMode, by_value=True, missing=None, data_key="inventoryMode"
    )
    origin = marshmallow_enum.EnumField(types.CartOrigin, by_value=True)
    tax_calculation_mode = marshmallow_enum.EnumField(
        types.TaxCalculationMode,
        by_value=True,
        missing=None,
        data_key="taxCalculationMode",
    )
    shipping_rate_input = helpers.Discriminator(
        discriminator_field=("type", "type"),
        discriminator_schemas={
            "Classification": "commercetools._schemas._cart.ClassificationShippingRateInputSchema",
            "Score": "commercetools._schemas._cart.ScoreShippingRateInputSchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingRateInput",
    )
    item_shipping_addresses = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
        data_key="itemShippingAddresses",
    )
    refused_gifts = helpers.LazyNestedField(
        nested="commercetools._schemas._cart_discount.CartDiscountReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        data_key="refusedGifts",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Order(**data)


class OrderUpdateActionSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.OrderUpdateAction`."""

    action = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderUpdateAction(**data)


class OrderUpdateSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.OrderUpdate`."""

    version = marshmallow.fields.Integer(allow_none=True)
    actions = marshmallow.fields.List(
        helpers.Discriminator(
            discriminator_field=("action", "action"),
            discriminator_schemas={
                "addDelivery": "commercetools._schemas._order.OrderAddDeliveryActionSchema",
                "addItemShippingAddress": "commercetools._schemas._order.OrderAddItemShippingAddressActionSchema",
                "addParcelToDelivery": "commercetools._schemas._order.OrderAddParcelToDeliveryActionSchema",
                "addPayment": "commercetools._schemas._order.OrderAddPaymentActionSchema",
                "addReturnInfo": "commercetools._schemas._order.OrderAddReturnInfoActionSchema",
                "changeOrderState": "commercetools._schemas._order.OrderChangeOrderStateActionSchema",
                "changePaymentState": "commercetools._schemas._order.OrderChangePaymentStateActionSchema",
                "changeShipmentState": "commercetools._schemas._order.OrderChangeShipmentStateActionSchema",
                "importCustomLineItemState": "commercetools._schemas._order.OrderImportCustomLineItemStateActionSchema",
                "importLineItemState": "commercetools._schemas._order.OrderImportLineItemStateActionSchema",
                "removeDelivery": "commercetools._schemas._order.OrderRemoveDeliveryActionSchema",
                "removeItemShippingAddress": "commercetools._schemas._order.OrderRemoveItemShippingAddressActionSchema",
                "removeParcelFromDelivery": "commercetools._schemas._order.OrderRemoveParcelFromDeliveryActionSchema",
                "removePayment": "commercetools._schemas._order.OrderRemovePaymentActionSchema",
                "setBillingAddress": "commercetools._schemas._order.OrderSetBillingAddressActionSchema",
                "setCustomField": "commercetools._schemas._order.OrderSetCustomFieldActionSchema",
                "setCustomLineItemCustomField": "commercetools._schemas._order.OrderSetCustomLineItemCustomFieldActionSchema",
                "setCustomLineItemCustomType": "commercetools._schemas._order.OrderSetCustomLineItemCustomTypeActionSchema",
                "setCustomLineItemShippingDetails": "commercetools._schemas._order.OrderSetCustomLineItemShippingDetailsActionSchema",
                "setCustomType": "commercetools._schemas._order.OrderSetCustomTypeActionSchema",
                "setCustomerEmail": "commercetools._schemas._order.OrderSetCustomerEmailActionSchema",
                "setCustomerId": "commercetools._schemas._order.OrderSetCustomerIdActionSchema",
                "setDeliveryAddress": "commercetools._schemas._order.OrderSetDeliveryAddressActionSchema",
                "setDeliveryItems": "commercetools._schemas._order.OrderSetDeliveryItemsActionSchema",
                "setLineItemCustomField": "commercetools._schemas._order.OrderSetLineItemCustomFieldActionSchema",
                "setLineItemCustomType": "commercetools._schemas._order.OrderSetLineItemCustomTypeActionSchema",
                "setLineItemShippingDetails": "commercetools._schemas._order.OrderSetLineItemShippingDetailsActionSchema",
                "setLocale": "commercetools._schemas._order.OrderSetLocaleActionSchema",
                "setOrderNumber": "commercetools._schemas._order.OrderSetOrderNumberActionSchema",
                "setParcelItems": "commercetools._schemas._order.OrderSetParcelItemsActionSchema",
                "setParcelMeasurements": "commercetools._schemas._order.OrderSetParcelMeasurementsActionSchema",
                "setParcelTrackingData": "commercetools._schemas._order.OrderSetParcelTrackingDataActionSchema",
                "setReturnPaymentState": "commercetools._schemas._order.OrderSetReturnPaymentStateActionSchema",
                "setReturnShipmentState": "commercetools._schemas._order.OrderSetReturnShipmentStateActionSchema",
                "setShippingAddress": "commercetools._schemas._order.OrderSetShippingAddressActionSchema",
                "setStore": "commercetools._schemas._order.OrderSetStoreActionSchema",
                "transitionCustomLineItemState": "commercetools._schemas._order.OrderTransitionCustomLineItemStateActionSchema",
                "transitionLineItemState": "commercetools._schemas._order.OrderTransitionLineItemStateActionSchema",
                "transitionState": "commercetools._schemas._order.OrderTransitionStateActionSchema",
                "updateItemShippingAddress": "commercetools._schemas._order.OrderUpdateItemShippingAddressActionSchema",
                "updateSyncInfo": "commercetools._schemas._order.OrderUpdateSyncInfoActionSchema",
            },
            unknown=marshmallow.EXCLUDE,
            allow_none=True,
        ),
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.OrderUpdate(**data)


class ParcelDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ParcelDraft`."""

    measurements = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ParcelMeasurementsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    tracking_data = helpers.LazyNestedField(
        nested="commercetools._schemas._order.TrackingDataSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="trackingData",
    )
    items = helpers.LazyNestedField(
        nested="commercetools._schemas._order.DeliveryItemSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ParcelDraft(**data)


class ParcelMeasurementsSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ParcelMeasurements`."""

    height_in_millimeter = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="heightInMillimeter"
    )
    length_in_millimeter = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="lengthInMillimeter"
    )
    width_in_millimeter = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="widthInMillimeter"
    )
    weight_in_gram = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="weightInGram"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ParcelMeasurements(**data)


class ParcelSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Parcel`."""

    id = marshmallow.fields.String(allow_none=True)
    created_at = marshmallow.fields.DateTime(allow_none=True, data_key="createdAt")
    measurements = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ParcelMeasurementsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    tracking_data = helpers.LazyNestedField(
        nested="commercetools._schemas._order.TrackingDataSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="trackingData",
    )
    items = helpers.LazyNestedField(
        nested="commercetools._schemas._order.DeliveryItemSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Parcel(**data)


class PaymentInfoSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.PaymentInfo`."""

    payments = helpers.LazyNestedField(
        nested="commercetools._schemas._payment.PaymentReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.PaymentInfo(**data)


class ProductVariantImportDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProductVariantImportDraft`."""

    id = marshmallow.fields.Integer(allow_none=True, missing=None)
    sku = marshmallow.fields.String(allow_none=True, missing=None)
    prices = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    attributes = helpers.LazyNestedField(
        nested="commercetools._schemas._product.AttributeSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    images = helpers.LazyNestedField(
        nested="commercetools._schemas._common.ImageSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProductVariantImportDraft(**data)


class ReturnInfoSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ReturnInfo`."""

    items = marshmallow.fields.List(
        helpers.Discriminator(
            discriminator_field=("type", "type"),
            discriminator_schemas={
                "CustomLineItemReturnItem": "commercetools._schemas._order.CustomLineItemReturnItemSchema",
                "LineItemReturnItem": "commercetools._schemas._order.LineItemReturnItemSchema",
            },
            unknown=marshmallow.EXCLUDE,
            allow_none=True,
        )
    )
    return_tracking_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="returnTrackingId"
    )
    return_date = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="returnDate"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ReturnInfo(**data)


class ReturnItemDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ReturnItemDraft`."""

    quantity = marshmallow.fields.Integer(allow_none=True)
    line_item_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="lineItemId"
    )
    custom_line_item_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="customLineItemId"
    )
    comment = marshmallow.fields.String(allow_none=True, missing=None)
    shipment_state = marshmallow_enum.EnumField(
        types.ReturnShipmentState, by_value=True, data_key="shipmentState"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ReturnItemDraft(**data)


class ReturnItemSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ReturnItem`."""

    id = marshmallow.fields.String(allow_none=True)
    quantity = marshmallow.fields.Integer(allow_none=True)
    type = marshmallow.fields.String(allow_none=True)
    comment = marshmallow.fields.String(allow_none=True, missing=None)
    shipment_state = marshmallow_enum.EnumField(
        types.ReturnShipmentState, by_value=True, data_key="shipmentState"
    )
    payment_state = marshmallow_enum.EnumField(
        types.ReturnPaymentState, by_value=True, data_key="paymentState"
    )
    last_modified_at = marshmallow.fields.DateTime(
        allow_none=True, data_key="lastModifiedAt"
    )
    created_at = marshmallow.fields.DateTime(allow_none=True, data_key="createdAt")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.ReturnItem(**data)


class ShippingInfoImportDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ShippingInfoImportDraft`."""

    shipping_method_name = marshmallow.fields.String(
        allow_none=True, data_key="shippingMethodName"
    )
    price = helpers.LazyNestedField(
        nested="commercetools._schemas._common.MoneySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    shipping_rate = helpers.LazyNestedField(
        nested="commercetools._schemas._shipping_method.ShippingRateDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="shippingRate",
    )
    tax_rate = helpers.LazyNestedField(
        nested="commercetools._schemas._tax_category.TaxRateSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="taxRate",
    )
    tax_category = helpers.LazyNestedField(
        nested="commercetools._schemas._tax_category.TaxCategoryResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="taxCategory",
    )
    shipping_method = helpers.LazyNestedField(
        nested="commercetools._schemas._shipping_method.ShippingMethodResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingMethod",
    )
    deliveries = helpers.LazyNestedField(
        nested="commercetools._schemas._order.DeliverySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    discounted_price = helpers.LazyNestedField(
        nested="commercetools._schemas._order.DiscountedLineItemPriceDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="discountedPrice",
    )
    shipping_method_state = marshmallow_enum.EnumField(
        types.ShippingMethodState,
        by_value=True,
        missing=None,
        data_key="shippingMethodState",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ShippingInfoImportDraft(**data)


class StagedOrderUpdateActionSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.StagedOrderUpdateAction`."""

    action = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.StagedOrderUpdateAction(**data)


class SyncInfoSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.SyncInfo`."""

    channel = helpers.LazyNestedField(
        nested="commercetools._schemas._channel.ChannelReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    external_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="externalId"
    )
    synced_at = marshmallow.fields.DateTime(allow_none=True, data_key="syncedAt")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.SyncInfo(**data)


class TaxedItemPriceDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.TaxedItemPriceDraft`."""

    total_net = helpers.LazyNestedField(
        nested="commercetools._schemas._common.MoneySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="totalNet",
    )
    total_gross = helpers.LazyNestedField(
        nested="commercetools._schemas._common.MoneySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="totalGross",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.TaxedItemPriceDraft(**data)


class TrackingDataSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.TrackingData`."""

    tracking_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="trackingId"
    )
    carrier = marshmallow.fields.String(allow_none=True, missing=None)
    provider = marshmallow.fields.String(allow_none=True, missing=None)
    provider_transaction = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="providerTransaction"
    )
    is_return = marshmallow.fields.Bool(
        allow_none=True, missing=None, data_key="isReturn"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.TrackingData(**data)


class CustomLineItemReturnItemSchema(ReturnItemSchema):
    """Marshmallow schema for :class:`commercetools.types.CustomLineItemReturnItem`."""

    custom_line_item_id = marshmallow.fields.String(
        allow_none=True, data_key="customLineItemId"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.CustomLineItemReturnItem(**data)


class LineItemReturnItemSchema(ReturnItemSchema):
    """Marshmallow schema for :class:`commercetools.types.LineItemReturnItem`."""

    line_item_id = marshmallow.fields.String(allow_none=True, data_key="lineItemId")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.LineItemReturnItem(**data)


class OrderAddDeliveryActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderAddDeliveryAction`."""

    items = helpers.LazyNestedField(
        nested="commercetools._schemas._order.DeliveryItemSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    parcels = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ParcelDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderAddDeliveryAction(**data)


class OrderAddItemShippingAddressActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderAddItemShippingAddressAction`."""

    address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderAddItemShippingAddressAction(**data)


class OrderAddParcelToDeliveryActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderAddParcelToDeliveryAction`."""

    delivery_id = marshmallow.fields.String(allow_none=True, data_key="deliveryId")
    measurements = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ParcelMeasurementsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    tracking_data = helpers.LazyNestedField(
        nested="commercetools._schemas._order.TrackingDataSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="trackingData",
    )
    items = helpers.LazyNestedField(
        nested="commercetools._schemas._order.DeliveryItemSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderAddParcelToDeliveryAction(**data)


class OrderAddPaymentActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderAddPaymentAction`."""

    payment = helpers.LazyNestedField(
        nested="commercetools._schemas._payment.PaymentResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderAddPaymentAction(**data)


class OrderAddReturnInfoActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderAddReturnInfoAction`."""

    return_tracking_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="returnTrackingId"
    )
    items = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ReturnItemDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    return_date = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="returnDate"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderAddReturnInfoAction(**data)


class OrderChangeOrderStateActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderChangeOrderStateAction`."""

    order_state = marshmallow_enum.EnumField(
        types.OrderState, by_value=True, data_key="orderState"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderChangeOrderStateAction(**data)


class OrderChangePaymentStateActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderChangePaymentStateAction`."""

    payment_state = marshmallow_enum.EnumField(
        types.PaymentState, by_value=True, missing=None, data_key="paymentState"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderChangePaymentStateAction(**data)


class OrderChangeShipmentStateActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderChangeShipmentStateAction`."""

    shipment_state = marshmallow_enum.EnumField(
        types.ShipmentState, by_value=True, missing=None, data_key="shipmentState"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderChangeShipmentStateAction(**data)


class OrderImportCustomLineItemStateActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderImportCustomLineItemStateAction`."""

    custom_line_item_id = marshmallow.fields.String(
        allow_none=True, data_key="customLineItemId"
    )
    state = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ItemStateSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderImportCustomLineItemStateAction(**data)


class OrderImportLineItemStateActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderImportLineItemStateAction`."""

    line_item_id = marshmallow.fields.String(allow_none=True, data_key="lineItemId")
    state = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ItemStateSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderImportLineItemStateAction(**data)


class OrderRemoveDeliveryActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderRemoveDeliveryAction`."""

    delivery_id = marshmallow.fields.String(allow_none=True, data_key="deliveryId")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderRemoveDeliveryAction(**data)


class OrderRemoveItemShippingAddressActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderRemoveItemShippingAddressAction`."""

    address_key = marshmallow.fields.String(allow_none=True, data_key="addressKey")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderRemoveItemShippingAddressAction(**data)


class OrderRemoveParcelFromDeliveryActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderRemoveParcelFromDeliveryAction`."""

    parcel_id = marshmallow.fields.String(allow_none=True, data_key="parcelId")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderRemoveParcelFromDeliveryAction(**data)


class OrderRemovePaymentActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderRemovePaymentAction`."""

    payment = helpers.LazyNestedField(
        nested="commercetools._schemas._payment.PaymentResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderRemovePaymentAction(**data)


class OrderSetBillingAddressActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetBillingAddressAction`."""

    address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetBillingAddressAction(**data)


class OrderSetCustomFieldActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetCustomFieldAction`."""

    name = marshmallow.fields.String(allow_none=True)
    value = marshmallow.fields.Raw(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetCustomFieldAction(**data)


class OrderSetCustomLineItemCustomFieldActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetCustomLineItemCustomFieldAction`."""

    custom_line_item_id = marshmallow.fields.String(
        allow_none=True, data_key="customLineItemId"
    )
    name = marshmallow.fields.String(allow_none=True)
    value = marshmallow.fields.Raw(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetCustomLineItemCustomFieldAction(**data)


class OrderSetCustomLineItemCustomTypeActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetCustomLineItemCustomTypeAction`."""

    custom_line_item_id = marshmallow.fields.String(
        allow_none=True, data_key="customLineItemId"
    )
    type = helpers.LazyNestedField(
        nested="commercetools._schemas._type.TypeResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    fields = FieldContainerField(allow_none=True, missing=None)  # type: ignore

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetCustomLineItemCustomTypeAction(**data)


class OrderSetCustomLineItemShippingDetailsActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetCustomLineItemShippingDetailsAction`."""

    custom_line_item_id = marshmallow.fields.String(
        allow_none=True, data_key="customLineItemId"
    )
    shipping_details = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.ItemShippingDetailsDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingDetails",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetCustomLineItemShippingDetailsAction(**data)


class OrderSetCustomTypeActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetCustomTypeAction`."""

    type = helpers.LazyNestedField(
        nested="commercetools._schemas._type.TypeResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    fields = FieldContainerField(allow_none=True, missing=None)  # type: ignore

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetCustomTypeAction(**data)


class OrderSetCustomerEmailActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetCustomerEmailAction`."""

    email = marshmallow.fields.String(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetCustomerEmailAction(**data)


class OrderSetCustomerIdActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetCustomerIdAction`."""

    customer_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="customerId"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetCustomerIdAction(**data)


class OrderSetDeliveryAddressActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetDeliveryAddressAction`."""

    delivery_id = marshmallow.fields.String(allow_none=True, data_key="deliveryId")
    address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetDeliveryAddressAction(**data)


class OrderSetDeliveryItemsActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetDeliveryItemsAction`."""

    delivery_id = marshmallow.fields.String(allow_none=True, data_key="deliveryId")
    items = helpers.LazyNestedField(
        nested="commercetools._schemas._order.DeliveryItemSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetDeliveryItemsAction(**data)


class OrderSetLineItemCustomFieldActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetLineItemCustomFieldAction`."""

    line_item_id = marshmallow.fields.String(allow_none=True, data_key="lineItemId")
    name = marshmallow.fields.String(allow_none=True)
    value = marshmallow.fields.Raw(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetLineItemCustomFieldAction(**data)


class OrderSetLineItemCustomTypeActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetLineItemCustomTypeAction`."""

    line_item_id = marshmallow.fields.String(allow_none=True, data_key="lineItemId")
    type = helpers.LazyNestedField(
        nested="commercetools._schemas._type.TypeResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    fields = FieldContainerField(allow_none=True, missing=None)  # type: ignore

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetLineItemCustomTypeAction(**data)


class OrderSetLineItemShippingDetailsActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetLineItemShippingDetailsAction`."""

    line_item_id = marshmallow.fields.String(allow_none=True, data_key="lineItemId")
    shipping_details = helpers.LazyNestedField(
        nested="commercetools._schemas._cart.ItemShippingDetailsDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingDetails",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetLineItemShippingDetailsAction(**data)


class OrderSetLocaleActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetLocaleAction`."""

    locale = marshmallow.fields.String(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetLocaleAction(**data)


class OrderSetOrderNumberActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetOrderNumberAction`."""

    order_number = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="orderNumber"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetOrderNumberAction(**data)


class OrderSetParcelItemsActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetParcelItemsAction`."""

    parcel_id = marshmallow.fields.String(allow_none=True, data_key="parcelId")
    items = helpers.LazyNestedField(
        nested="commercetools._schemas._order.DeliveryItemSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetParcelItemsAction(**data)


class OrderSetParcelMeasurementsActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetParcelMeasurementsAction`."""

    parcel_id = marshmallow.fields.String(allow_none=True, data_key="parcelId")
    measurements = helpers.LazyNestedField(
        nested="commercetools._schemas._order.ParcelMeasurementsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetParcelMeasurementsAction(**data)


class OrderSetParcelTrackingDataActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetParcelTrackingDataAction`."""

    parcel_id = marshmallow.fields.String(allow_none=True, data_key="parcelId")
    tracking_data = helpers.LazyNestedField(
        nested="commercetools._schemas._order.TrackingDataSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="trackingData",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetParcelTrackingDataAction(**data)


class OrderSetReturnPaymentStateActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetReturnPaymentStateAction`."""

    return_item_id = marshmallow.fields.String(allow_none=True, data_key="returnItemId")
    payment_state = marshmallow_enum.EnumField(
        types.ReturnPaymentState, by_value=True, data_key="paymentState"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetReturnPaymentStateAction(**data)


class OrderSetReturnShipmentStateActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetReturnShipmentStateAction`."""

    return_item_id = marshmallow.fields.String(allow_none=True, data_key="returnItemId")
    shipment_state = marshmallow_enum.EnumField(
        types.ReturnShipmentState, by_value=True, data_key="shipmentState"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetReturnShipmentStateAction(**data)


class OrderSetShippingAddressActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetShippingAddressAction`."""

    address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetShippingAddressAction(**data)


class OrderSetStoreActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderSetStoreAction`."""

    store = helpers.LazyNestedField(
        nested="commercetools._schemas._store.StoreResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderSetStoreAction(**data)


class OrderTransitionCustomLineItemStateActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderTransitionCustomLineItemStateAction`."""

    custom_line_item_id = marshmallow.fields.String(
        allow_none=True, data_key="customLineItemId"
    )
    quantity = marshmallow.fields.Integer(allow_none=True)
    from_state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="fromState",
    )
    to_state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="toState",
    )
    actual_transition_date = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="actualTransitionDate"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderTransitionCustomLineItemStateAction(**data)


class OrderTransitionLineItemStateActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderTransitionLineItemStateAction`."""

    line_item_id = marshmallow.fields.String(allow_none=True, data_key="lineItemId")
    quantity = marshmallow.fields.Integer(allow_none=True)
    from_state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="fromState",
    )
    to_state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="toState",
    )
    actual_transition_date = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="actualTransitionDate"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderTransitionLineItemStateAction(**data)


class OrderTransitionStateActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderTransitionStateAction`."""

    state = helpers.LazyNestedField(
        nested="commercetools._schemas._state.StateResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    force = marshmallow.fields.Bool(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderTransitionStateAction(**data)


class OrderUpdateItemShippingAddressActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderUpdateItemShippingAddressAction`."""

    address = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AddressSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderUpdateItemShippingAddressAction(**data)


class OrderUpdateSyncInfoActionSchema(OrderUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.OrderUpdateSyncInfoAction`."""

    channel = helpers.LazyNestedField(
        nested="commercetools._schemas._channel.ChannelResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    external_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="externalId"
    )
    synced_at = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="syncedAt"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.OrderUpdateSyncInfoAction(**data)



import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._product_discount import (
    ProductDiscountDraftSchema,
    ProductDiscountPagedQueryResponseSchema,
    ProductDiscountSchema,
    ProductDiscountUpdateSchema,
)
from commercetools.testing.abstract import BaseModel, ServiceBackend


class ProductDiscountsModel(BaseModel):
    _primary_type_name = "product-discount"
    _resource_schema = ProductDiscountSchema

    def _create_from_draft(
        self, draft: types.ProductDiscountDraft, id: typing.Optional[str] = None
    ) -> types.ProductDiscount:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        return types.ProductDiscount(
            id=str(object_id),
            version=1,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
            name=draft.name,
            description=draft.description,
            value=draft.value,
            predicate=draft.predicate,
            sort_order=draft.sort_order,
            references=[],
            is_active=draft.is_active,
            valid_from=draft.valid_from,
            valid_until=draft.valid_until,
        )


class ProductDiscountsBackend(ServiceBackend):
    service_path = "product-discounts"
    model_class = ProductDiscountsModel

    _schema_draft = ProductDiscountDraftSchema
    _schema_update = ProductDiscountUpdateSchema
    _schema_query_response = ProductDiscountPagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
        ]



import uuid

import pytest

from commercetools import types


def test_cart_get_by_id(client, cart_draft):
    cart = client.carts.create(cart_draft)

    assert cart.id


@pytest.fixture
def cart_draft(client):
    product_1 = client.products.create(
        types.ProductDraft(
            key=f"product-1",
            product_type=types.ProductTypeResourceIdentifier(key="dummy"),
            name=types.LocalizedString(en=f"my-product-1"),
            slug=types.LocalizedString(en=f"my-product-1"),
            publish=True,
        )
    )
    product_2 = client.products.create(
        types.ProductDraft(
            key=f"product-2",
            product_type=types.ProductTypeResourceIdentifier(key="dummy"),
            name=types.LocalizedString(en=f"my-product-2"),
            slug=types.LocalizedString(en=f"my-product-2"),
            publish=True,
        )
    )

    return types.CartDraft(
        customer_id=str(uuid.uuid4()),
        customer_email="foo@example.com",
        currency="GBP",
        anonymous_id=str(uuid.uuid4()),
        country="GB",
        inventory_mode=types.InventoryMode.NONE,
        tax_mode=types.TaxMode.PLATFORM,
        tax_rounding_mode=types.RoundingMode.HALF_EVEN,
        tax_calculation_mode=types.TaxCalculationMode.LINE_ITEM_LEVEL,
        line_items=[
            types.LineItemDraft(product_id=product_1.id, quantity=1),
            types.LineItemDraft(product_id=product_2.id, quantity=2),
        ],
        locale="en",
        origin=types.CartOrigin.CUSTOMER,
    )


def test_update_actions(commercetools_api, client, cart_draft):
    cart = client.carts.create(cart_draft)
    payment_reference = types.PaymentReference(id=str(uuid.uuid4()))
    cart = client.carts.update_by_id(
        cart.id,
        cart.version,
        actions=[types.CartAddPaymentAction(payment=payment_reference)],
    )

    type_draft = types.TypeDraft(
        key="foobar",
        resource_type_ids=[types.ResourceTypeId.ORDER],
        name={"en-US": "test"},
        field_definitions=[
            types.FieldDefinition(
                type=types.CustomFieldStringType(),
                name="foo1",
                label={"en-US": "foo-1"},
                required=False,
            ),
            types.FieldDefinition(
                type=types.CustomFieldSetType(element_type=None),
                name="foo2",
                label={"en-US": "foo-2"},
                required=False,
            ),
            types.FieldDefinition(
                type=types.CustomFieldBooleanType(),
                name="foo3",
                label={"en-US": "foo-3"},
                required=False,
            ),
        ],
    )
    custom_type = client.types.create(type_draft)
    assert custom_type.id

    assert cart.payment_info.payments[0] == payment_reference
    cart = client.carts.update_by_id(
        cart.id,
        cart.version,
        actions=[
            types.CartSetCustomTypeAction(
                type=types.TypeResourceIdentifier(id=custom_type.id)
            )
        ],
    )

    client.carts.update_by_id(
        cart.id,
        cart.version,
        actions=[
            types.CartSetCustomFieldAction(name="foo1", value="bar"),
            types.CartSetCustomFieldAction(name="foo2", value=["bar"]),
            types.CartSetCustomFieldAction(name="foo3", value=False),
        ],
    )

    cart = client.carts.get_by_id(cart.id)

    assert all(key in cart.custom.fields for key in ["foo1", "foo2", "foo3"])
    assert cart.custom.fields["foo1"] == "bar"
    assert cart.custom.fields["foo2"] == ["bar"]
    assert cart.custom.fields["foo3"] is False



# DO NOT EDIT! This file is automatically generated
import typing

from commercetools._schemas._cart import CartPagedQueryResponseSchema
from commercetools._schemas._customer import CustomerSignInResultSchema
from commercetools._schemas._me import (
    MyCartDraftSchema,
    MyCartSchema,
    MyCustomerDraftSchema,
    MyCustomerSchema,
    MyOrderFromCartDraftSchema,
    MyOrderSchema,
    MyPaymentDraftSchema,
    MyPaymentPagedQueryResponseSchema,
    MyPaymentSchema,
    MyPaymentUpdateSchema,
    MyShoppingListDraftSchema,
    MyShoppingListUpdateSchema,
)
from commercetools._schemas._order import OrderPagedQueryResponseSchema
from commercetools._schemas._shopping_list import (
    MyShoppingListSchema,
    ShoppingListPagedQueryResponseSchema,
)
from commercetools.helpers import RemoveEmptyValuesMixin
from commercetools.types._cart import CartPagedQueryResponse
from commercetools.types._customer import CustomerSignInResult
from commercetools.types._me import (
    MyCart,
    MyCartDraft,
    MyCustomer,
    MyCustomerDraft,
    MyOrder,
    MyOrderFromCartDraft,
    MyPayment,
    MyPaymentDraft,
    MyPaymentPagedQueryResponse,
    MyPaymentUpdate,
    MyPaymentUpdateAction,
    MyShoppingListDraft,
    MyShoppingListUpdate,
    MyShoppingListUpdateAction,
)
from commercetools.types._order import OrderPagedQueryResponse
from commercetools.types._shopping_list import (
    MyShoppingList,
    ShoppingListPagedQueryResponse,
)
from commercetools.typing import OptionalListStr

from . import abstract, traits


class _MeQuerySchema(
    traits.SortableSchema,
    traits.PagingSchema,
    traits.ExpandableSchema,
    traits.QuerySchema,
):
    pass


class _MeUpdateSchema(traits.ExpandableSchema, traits.VersionedSchema):
    pass


class _MeDeleteSchema(traits.VersionedSchema, traits.ExpandableSchema):
    pass


class MeService(abstract.AbstractService):
    def my_cart_get_by_id(self, id: str, *, expand: OptionalListStr = None) -> MyCart:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"me/carts/{id}", params=params, schema_cls=MyCartSchema
        )

    def my_order_get_by_id(self, id: str, *, expand: OptionalListStr = None) -> MyOrder:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"me/orders/{id}", params=params, schema_cls=MyOrderSchema
        )

    def my_payment_get_by_id(
        self, id: str, *, expand: OptionalListStr = None
    ) -> MyPayment:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"me/payments/{id}", params=params, schema_cls=MyPaymentSchema
        )

    def my_payment_get_by_key(
        self, key: str, *, expand: OptionalListStr = None
    ) -> MyPayment:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"me/payments/key={key}", params=params, schema_cls=MyPaymentSchema
        )

    def my_shopping_list_get_by_id(
        self, id: str, *, expand: OptionalListStr = None
    ) -> MyShoppingList:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"me/shopping-lists/{id}",
            params=params,
            schema_cls=MyShoppingListSchema,
        )

    def my_shopping_list_get_by_key(
        self, key: str, *, expand: OptionalListStr = None
    ) -> MyShoppingList:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"me/shopping-lists/key={key}",
            params=params,
            schema_cls=MyShoppingListSchema,
        )

    def my_cart_query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> CartPagedQueryResponse:
        """A shopping cart holds product variants and can be ordered.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _MeQuerySchema,
        )
        return self._client._get(
            endpoint="me/carts", params=params, schema_cls=CartPagedQueryResponseSchema
        )

    def my_order_query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> OrderPagedQueryResponse:
        """An order can be created from a cart, usually after a checkout process has
        been completed.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _MeQuerySchema,
        )
        return self._client._get(
            endpoint="me/orders",
            params=params,
            schema_cls=OrderPagedQueryResponseSchema,
        )

    def my_payment_query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> MyPaymentPagedQueryResponse:
        """The My Payments endpoint creates and provides access to payments scoped
        to a specific user.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _MeQuerySchema,
        )
        return self._client._get(
            endpoint="me/payments",
            params=params,
            schema_cls=MyPaymentPagedQueryResponseSchema,
        )

    def my_shopping_list_query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> ShoppingListPagedQueryResponse:
        """The My Shopping Lists endpoint creates and provides access to shopping
        lists scoped to a specific user.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _MeQuerySchema,
        )
        return self._client._get(
            endpoint="me/shopping-lists",
            params=params,
            schema_cls=ShoppingListPagedQueryResponseSchema,
        )

    def query(
        self,
        *,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        expand: OptionalListStr = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> MyCustomer:
        params = self._serialize_params(
            {
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "expand": expand,
                "where": where,
                "predicate_var": predicate_var,
            },
            _MeQuerySchema,
        )
        return self._client._get(
            endpoint="me", params=params, schema_cls=MyCustomerSchema
        )

    def create(self) -> MyCustomer:
        """Create a customer"""
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="me",
            params=params,
            data_object=None,
            request_schema_cls=None,
            response_schema_cls=MyCustomerSchema,
        )

    def my_cart_create(
        self, draft: MyCartDraft, *, expand: OptionalListStr = None
    ) -> MyCart:
        """A shopping cart holds product variants and can be ordered.
        """
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._post(
            endpoint="me/carts",
            params=params,
            data_object=draft,
            request_schema_cls=MyCartDraftSchema,
            response_schema_cls=MyCartSchema,
        )

    def my_order_create(
        self, draft: MyOrderFromCartDraft, *, expand: OptionalListStr = None
    ) -> MyOrder:
        """An order can be created from a cart, usually after a checkout process has
        been completed.
        """
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._post(
            endpoint="me/orders",
            params=params,
            data_object=draft,
            request_schema_cls=MyOrderFromCartDraftSchema,
            response_schema_cls=MyOrderSchema,
        )

    def my_payment_create(
        self, draft: MyPaymentDraft, *, expand: OptionalListStr = None
    ) -> MyPayment:
        """The My Payments endpoint creates and provides access to payments scoped
        to a specific user.
        """
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._post(
            endpoint="me/payments",
            params=params,
            data_object=draft,
            request_schema_cls=MyPaymentDraftSchema,
            response_schema_cls=MyPaymentSchema,
        )

    def my_shopping_list_create(
        self, draft: MyShoppingListDraft, *, expand: OptionalListStr = None
    ) -> MyShoppingList:
        """The My Shopping Lists endpoint creates and provides access to shopping
        lists scoped to a specific user.
        """
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._post(
            endpoint="me/shopping-lists",
            params=params,
            data_object=draft,
            request_schema_cls=MyShoppingListDraftSchema,
            response_schema_cls=MyShoppingListSchema,
        )

    def my_payment_update_by_id(
        self,
        id: str,
        version: int,
        actions: typing.List[MyPaymentUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> MyPayment:
        params = self._serialize_params({"expand": expand}, _MeUpdateSchema)
        update_action = MyPaymentUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"me/payments/{id}",
            params=params,
            data_object=update_action,
            request_schema_cls=MyPaymentUpdateSchema,
            response_schema_cls=MyPaymentSchema,
            force_update=force_update,
        )

    def my_payment_update_by_key(
        self,
        key: str,
        version: int,
        actions: typing.List[MyPaymentUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> MyPayment:
        params = self._serialize_params({"expand": expand}, _MeUpdateSchema)
        update_action = MyPaymentUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"me/payments/key={key}",
            params=params,
            data_object=update_action,
            request_schema_cls=MyPaymentUpdateSchema,
            response_schema_cls=MyPaymentSchema,
            force_update=force_update,
        )

    def my_shopping_list_update_by_id(
        self,
        id: str,
        version: int,
        actions: typing.List[MyShoppingListUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> MyShoppingList:
        params = self._serialize_params({"expand": expand}, _MeUpdateSchema)
        update_action = MyShoppingListUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"me/shopping-lists/{id}",
            params=params,
            data_object=update_action,
            request_schema_cls=MyShoppingListUpdateSchema,
            response_schema_cls=MyShoppingListSchema,
            force_update=force_update,
        )

    def my_shopping_list_update_by_key(
        self,
        key: str,
        version: int,
        actions: typing.List[MyShoppingListUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> MyShoppingList:
        params = self._serialize_params({"expand": expand}, _MeUpdateSchema)
        update_action = MyShoppingListUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"me/shopping-lists/key={key}",
            params=params,
            data_object=update_action,
            request_schema_cls=MyShoppingListUpdateSchema,
            response_schema_cls=MyShoppingListSchema,
            force_update=force_update,
        )

    def my_cart_delete_by_id(
        self,
        id: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> MyCart:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _MeDeleteSchema
        )
        return self._client._delete(
            endpoint=f"me/carts/{id}",
            params=params,
            response_schema_cls=MyCartSchema,
            force_delete=force_delete,
        )

    def my_payment_delete_by_id(
        self,
        id: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> MyPayment:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _MeDeleteSchema
        )
        return self._client._delete(
            endpoint=f"me/payments/{id}",
            params=params,
            response_schema_cls=MyPaymentSchema,
            force_delete=force_delete,
        )

    def my_payment_delete_by_key(
        self,
        key: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> MyPayment:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _MeDeleteSchema
        )
        return self._client._delete(
            endpoint=f"me/payments/key={key}",
            params=params,
            response_schema_cls=MyPaymentSchema,
            force_delete=force_delete,
        )

    def my_shopping_list_delete_by_id(
        self,
        id: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> MyShoppingList:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _MeDeleteSchema
        )
        return self._client._delete(
            endpoint=f"me/shopping-lists/{id}",
            params=params,
            response_schema_cls=MyShoppingListSchema,
            force_delete=force_delete,
        )

    def my_shopping_list_delete_by_key(
        self,
        key: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> MyShoppingList:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _MeDeleteSchema
        )
        return self._client._delete(
            endpoint=f"me/shopping-lists/key={key}",
            params=params,
            response_schema_cls=MyShoppingListSchema,
            force_delete=force_delete,
        )

    def active_cart(self) -> MyCart:
        params: typing.Dict[str, str] = {}
        return self._client._get(
            endpoint="me/active-cart", params=params, schema_cls=MyCartSchema
        )

    def email_confirm(self):
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="me/email/confirm",
            params=params,
            data_object=None,
            request_schema_cls=None,
        )

    def login(self) -> CustomerSignInResult:
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="me/login",
            params=params,
            data_object=None,
            request_schema_cls=None,
            response_schema_cls=CustomerSignInResultSchema,
        )

    def password(self) -> MyCustomer:
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="me/password",
            params=params,
            data_object=None,
            request_schema_cls=None,
            response_schema_cls=MyCustomerSchema,
        )

    def signup(self, draft: MyCustomerDraft) -> CustomerSignInResult:
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="me/signup",
            params=params,
            data_object=draft,
            request_schema_cls=MyCustomerDraftSchema,
            response_schema_cls=CustomerSignInResultSchema,
        )



# DO NOT EDIT! This file is automatically generated
import datetime
import typing

from commercetools.types._abstract import _BaseType
from commercetools.types._common import (
    BaseResource,
    Reference,
    ReferenceTypeId,
    ResourceIdentifier,
)

if typing.TYPE_CHECKING:
    from ._common import CreatedBy, LastModifiedBy
__all__ = [
    "SubRate",
    "TaxCategory",
    "TaxCategoryAddTaxRateAction",
    "TaxCategoryChangeNameAction",
    "TaxCategoryDraft",
    "TaxCategoryPagedQueryResponse",
    "TaxCategoryReference",
    "TaxCategoryRemoveTaxRateAction",
    "TaxCategoryReplaceTaxRateAction",
    "TaxCategoryResourceIdentifier",
    "TaxCategorySetDescriptionAction",
    "TaxCategorySetKeyAction",
    "TaxCategoryUpdate",
    "TaxCategoryUpdateAction",
    "TaxRate",
    "TaxRateDraft",
]


class SubRate(_BaseType):
    #: :class:`str`
    name: str
    #: :class:`float`
    amount: float

    def __init__(self, *, name: str, amount: float) -> None:
        self.name = name
        self.amount = amount
        super().__init__()

    def __repr__(self) -> str:
        return "SubRate(name=%r, amount=%r)" % (self.name, self.amount)


class TaxCategory(BaseResource):
    #: :class:`str`
    id: str
    #: :class:`int`
    version: int
    #: :class:`datetime.datetime` `(Named` ``createdAt`` `in Commercetools)`
    created_at: datetime.datetime
    #: :class:`datetime.datetime` `(Named` ``lastModifiedAt`` `in Commercetools)`
    last_modified_at: datetime.datetime
    #: Optional :class:`commercetools.types.LastModifiedBy` `(Named` ``lastModifiedBy`` `in Commercetools)`
    last_modified_by: typing.Optional["LastModifiedBy"]
    #: Optional :class:`commercetools.types.CreatedBy` `(Named` ``createdBy`` `in Commercetools)`
    created_by: typing.Optional["CreatedBy"]
    #: :class:`str`
    name: str
    #: Optional :class:`str`
    description: typing.Optional[str]
    #: List of :class:`commercetools.types.TaxRate`
    rates: typing.List["TaxRate"]
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(
        self,
        *,
        id: str,
        version: int,
        created_at: datetime.datetime,
        last_modified_at: datetime.datetime,
        name: str,
        rates: typing.List["TaxRate"],
        last_modified_by: typing.Optional["LastModifiedBy"] = None,
        created_by: typing.Optional["CreatedBy"] = None,
        description: typing.Optional[str] = None,
        key: typing.Optional[str] = None
    ) -> None:
        self.id = id
        self.version = version
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        self.last_modified_by = last_modified_by
        self.created_by = created_by
        self.name = name
        self.description = description
        self.rates = rates
        self.key = key
        super().__init__(
            id=id,
            version=version,
            created_at=created_at,
            last_modified_at=last_modified_at,
        )

    def __repr__(self) -> str:
        return (
            "TaxCategory(id=%r, version=%r, created_at=%r, last_modified_at=%r, last_modified_by=%r, created_by=%r, name=%r, description=%r, rates=%r, key=%r)"
            % (
                self.id,
                self.version,
                self.created_at,
                self.last_modified_at,
                self.last_modified_by,
                self.created_by,
                self.name,
                self.description,
                self.rates,
                self.key,
            )
        )


class TaxCategoryDraft(_BaseType):
    #: :class:`str`
    name: str
    #: Optional :class:`str`
    description: typing.Optional[str]
    #: List of :class:`commercetools.types.TaxRateDraft`
    rates: typing.List["TaxRateDraft"]
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(
        self,
        *,
        name: str,
        rates: typing.List["TaxRateDraft"],
        description: typing.Optional[str] = None,
        key: typing.Optional[str] = None
    ) -> None:
        self.name = name
        self.description = description
        self.rates = rates
        self.key = key
        super().__init__()

    def __repr__(self) -> str:
        return "TaxCategoryDraft(name=%r, description=%r, rates=%r, key=%r)" % (
            self.name,
            self.description,
            self.rates,
            self.key,
        )


class TaxCategoryPagedQueryResponse(_BaseType):
    #: :class:`int`
    limit: int
    #: :class:`int`
    count: int
    #: Optional :class:`int`
    total: typing.Optional[int]
    #: :class:`int`
    offset: int
    #: List of :class:`commercetools.types.TaxCategory`
    results: typing.Sequence["TaxCategory"]

    def __init__(
        self,
        *,
        limit: int,
        count: int,
        offset: int,
        results: typing.Sequence["TaxCategory"],
        total: typing.Optional[int] = None
    ) -> None:
        self.limit = limit
        self.count = count
        self.total = total
        self.offset = offset
        self.results = results
        super().__init__()

    def __repr__(self) -> str:
        return (
            "TaxCategoryPagedQueryResponse(limit=%r, count=%r, total=%r, offset=%r, results=%r)"
            % (self.limit, self.count, self.total, self.offset, self.results)
        )


class TaxCategoryReference(Reference):
    #: Optional :class:`commercetools.types.TaxCategory`
    obj: typing.Optional["TaxCategory"]

    def __init__(self, *, id: str, obj: typing.Optional["TaxCategory"] = None) -> None:
        self.obj = obj
        super().__init__(type_id=ReferenceTypeId.TAX_CATEGORY, id=id)

    def __repr__(self) -> str:
        return "TaxCategoryReference(type_id=%r, id=%r, obj=%r)" % (
            self.type_id,
            self.id,
            self.obj,
        )


class TaxCategoryResourceIdentifier(ResourceIdentifier):
    def __init__(
        self, *, id: typing.Optional[str] = None, key: typing.Optional[str] = None
    ) -> None:
        super().__init__(type_id=ReferenceTypeId.TAX_CATEGORY, id=id, key=key)

    def __repr__(self) -> str:
        return "TaxCategoryResourceIdentifier(type_id=%r, id=%r, key=%r)" % (
            self.type_id,
            self.id,
            self.key,
        )


class TaxCategoryUpdate(_BaseType):
    #: :class:`int`
    version: int
    #: :class:`list`
    actions: list

    def __init__(self, *, version: int, actions: list) -> None:
        self.version = version
        self.actions = actions
        super().__init__()

    def __repr__(self) -> str:
        return "TaxCategoryUpdate(version=%r, actions=%r)" % (
            self.version,
            self.actions,
        )


class TaxCategoryUpdateAction(_BaseType):
    #: :class:`str`
    action: str

    def __init__(self, *, action: str) -> None:
        self.action = action
        super().__init__()

    def __repr__(self) -> str:
        return "TaxCategoryUpdateAction(action=%r)" % (self.action,)


class TaxRate(_BaseType):
    #: Optional :class:`str`
    id: typing.Optional[str]
    #: :class:`str`
    name: str
    #: :class:`float`
    amount: float
    #: :class:`bool` `(Named` ``includedInPrice`` `in Commercetools)`
    included_in_price: bool
    #: :class:`str`
    country: "str"
    #: Optional :class:`str`
    state: typing.Optional[str]
    #: Optional list of :class:`commercetools.types.SubRate` `(Named` ``subRates`` `in Commercetools)`
    sub_rates: typing.Optional[typing.List["SubRate"]]

    def __init__(
        self,
        *,
        name: str,
        amount: float,
        included_in_price: bool,
        country: "str",
        id: typing.Optional[str] = None,
        state: typing.Optional[str] = None,
        sub_rates: typing.Optional[typing.List["SubRate"]] = None
    ) -> None:
        self.id = id
        self.name = name
        self.amount = amount
        self.included_in_price = included_in_price
        self.country = country
        self.state = state
        self.sub_rates = sub_rates
        super().__init__()

    def __repr__(self) -> str:
        return (
            "TaxRate(id=%r, name=%r, amount=%r, included_in_price=%r, country=%r, state=%r, sub_rates=%r)"
            % (
                self.id,
                self.name,
                self.amount,
                self.included_in_price,
                self.country,
                self.state,
                self.sub_rates,
            )
        )


class TaxRateDraft(_BaseType):
    #: :class:`str`
    name: str
    #: Optional :class:`float`
    amount: typing.Optional[float]
    #: :class:`bool` `(Named` ``includedInPrice`` `in Commercetools)`
    included_in_price: bool
    #: :class:`str`
    country: "str"
    #: Optional :class:`str`
    state: typing.Optional[str]
    #: Optional list of :class:`commercetools.types.SubRate` `(Named` ``subRates`` `in Commercetools)`
    sub_rates: typing.Optional[typing.List["SubRate"]]

    def __init__(
        self,
        *,
        name: str,
        included_in_price: bool,
        country: "str",
        amount: typing.Optional[float] = None,
        state: typing.Optional[str] = None,
        sub_rates: typing.Optional[typing.List["SubRate"]] = None
    ) -> None:
        self.name = name
        self.amount = amount
        self.included_in_price = included_in_price
        self.country = country
        self.state = state
        self.sub_rates = sub_rates
        super().__init__()

    def __repr__(self) -> str:
        return (
            "TaxRateDraft(name=%r, amount=%r, included_in_price=%r, country=%r, state=%r, sub_rates=%r)"
            % (
                self.name,
                self.amount,
                self.included_in_price,
                self.country,
                self.state,
                self.sub_rates,
            )
        )


class TaxCategoryAddTaxRateAction(TaxCategoryUpdateAction):
    #: :class:`commercetools.types.TaxRateDraft` `(Named` ``taxRate`` `in Commercetools)`
    tax_rate: "TaxRateDraft"

    def __init__(self, *, tax_rate: "TaxRateDraft") -> None:
        self.tax_rate = tax_rate
        super().__init__(action="addTaxRate")

    def __repr__(self) -> str:
        return "TaxCategoryAddTaxRateAction(action=%r, tax_rate=%r)" % (
            self.action,
            self.tax_rate,
        )


class TaxCategoryChangeNameAction(TaxCategoryUpdateAction):
    #: :class:`str`
    name: str

    def __init__(self, *, name: str) -> None:
        self.name = name
        super().__init__(action="changeName")

    def __repr__(self) -> str:
        return "TaxCategoryChangeNameAction(action=%r, name=%r)" % (
            self.action,
            self.name,
        )


class TaxCategoryRemoveTaxRateAction(TaxCategoryUpdateAction):
    #: :class:`str` `(Named` ``taxRateId`` `in Commercetools)`
    tax_rate_id: str

    def __init__(self, *, tax_rate_id: str) -> None:
        self.tax_rate_id = tax_rate_id
        super().__init__(action="removeTaxRate")

    def __repr__(self) -> str:
        return "TaxCategoryRemoveTaxRateAction(action=%r, tax_rate_id=%r)" % (
            self.action,
            self.tax_rate_id,
        )


class TaxCategoryReplaceTaxRateAction(TaxCategoryUpdateAction):
    #: :class:`str` `(Named` ``taxRateId`` `in Commercetools)`
    tax_rate_id: str
    #: :class:`commercetools.types.TaxRateDraft` `(Named` ``taxRate`` `in Commercetools)`
    tax_rate: "TaxRateDraft"

    def __init__(self, *, tax_rate_id: str, tax_rate: "TaxRateDraft") -> None:
        self.tax_rate_id = tax_rate_id
        self.tax_rate = tax_rate
        super().__init__(action="replaceTaxRate")

    def __repr__(self) -> str:
        return (
            "TaxCategoryReplaceTaxRateAction(action=%r, tax_rate_id=%r, tax_rate=%r)"
            % (self.action, self.tax_rate_id, self.tax_rate)
        )


class TaxCategorySetDescriptionAction(TaxCategoryUpdateAction):
    #: Optional :class:`str`
    description: typing.Optional[str]

    def __init__(self, *, description: typing.Optional[str] = None) -> None:
        self.description = description
        super().__init__(action="setDescription")

    def __repr__(self) -> str:
        return "TaxCategorySetDescriptionAction(action=%r, description=%r)" % (
            self.action,
            self.description,
        )


class TaxCategorySetKeyAction(TaxCategoryUpdateAction):
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(self, *, key: typing.Optional[str] = None) -> None:
        self.key = key
        super().__init__(action="setKey")

    def __repr__(self) -> str:
        return "TaxCategorySetKeyAction(action=%r, key=%r)" % (self.action, self.key)



import copy
import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._cart import CartSchema, ShippingInfoSchema
from commercetools._schemas._order import (
    DeliverySchema,
    OrderFromCartDraftSchema,
    OrderPagedQueryResponseSchema,
    OrderSchema,
    OrderUpdateSchema,
)
from commercetools._schemas._payment import PaymentReferenceSchema
from commercetools.testing.abstract import BaseModel, ServiceBackend
from commercetools.testing.utils import (
    set_custom_field,
    update_attribute,
    update_enum_attribute,
)
from commercetools.types import CartOrigin, OrderState


class OrdersModel(BaseModel):
    _primary_type_name = "order"
    _resource_schema = OrderSchema

    def _create_from_draft(
        self, draft: types.OrderFromCartDraft, id: typing.Optional[str] = None
    ) -> types.Order:
        """
        Note this implementation needs further refinement. For example:
         - Copying fields from an existing cart
         - Setting custom type fields?
        """

        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        cart_identifier = types.CartResourceIdentifier(id=draft.id)

        cart_data = self._storage.get_by_resource_identifier(cart_identifier)
        cart = None
        if cart_data:
            cart: types.Cart = CartSchema().load(cart_data)

        order = types.Order(
            id=str(object_id),
            version=1,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
            line_items=[],
            custom_line_items=[],
            total_price=copy.deepcopy(cart.total_price),
            sync_info=[],
            last_message_sequence_number=0,
            refused_gifts=[],
            order_number=draft.order_number,
            payment_state=draft.payment_state,
            order_state=OrderState.OPEN,
            origin=CartOrigin.CUSTOMER,
        )
        return order


def add_delivery():
    def updater(self, obj, action):
        parcels: typing.List[types.ParcelDraft] = getattr(action, "parcels")
        delivery = types.Delivery(
            id=str(uuid.uuid4()),
            created_at=datetime.datetime.now(datetime.timezone.utc),
            items=[],
            parcels=[
                types.Parcel(
                    id=str(uuid.uuid4()),
                    created_at=datetime.datetime.now(datetime.timezone.utc),
                    measurements=parcel_draft.measurements,
                    tracking_data=parcel_draft.tracking_data,
                    items=parcel_draft.items,
                )
                for parcel_draft in parcels
            ],
        )

        if not obj["shippingInfo"]:
            obj["shippingInfo"] = ShippingInfoSchema().dump(
                types.ShippingInfo(
                    shipping_method_name="dummy",
                    price=types.CentPrecisionMoney(
                        fraction_digits=0, cent_amount=0, currency_code="EUR"
                    ),
                    shipping_rate=types.ShippingRate(
                        price=types.CentPrecisionMoney(
                            fraction_digits=0, cent_amount=0, currency_code="EUR"
                        ),
                        tiers=[],
                    ),
                    shipping_method_state=types.ShippingMethodState(value=None),
                    deliveries=[],
                )
            )
        elif not obj["shippingInfo"]["deliveries"]:
            obj["shippingInfo"]["deliveries"] = []

        value = DeliverySchema().dump(delivery)
        if value not in obj["shippingInfo"]["deliveries"]:
            new = copy.deepcopy(obj)
            new["shippingInfo"]["deliveries"].append(value)
            return new
        return obj

    return updater


def add_payment():
    def updater(self, obj, action):
        obj = copy.deepcopy(obj)
        if "paymentInfo" in obj and obj["paymentInfo"]:
            payments = obj["paymentInfo"]["payments"]
        else:
            payments = []

        payments.append(PaymentReferenceSchema().dump(action.payment))
        obj["paymentInfo"] = {"payments": payments}
        return obj

    return updater


class OrdersBackend(ServiceBackend):
    service_path = "orders"
    model_class = OrdersModel
    _schema_draft = OrderFromCartDraftSchema
    _schema_update = OrderUpdateSchema
    _schema_query_response = OrderPagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^key=(?P<key>[^/]+)$", "GET", self.get_by_key),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^key=(?P<key>[^/]+)$", "POST", self.update_by_key),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
        ]

    _actions = {
        "changeOrderState": update_enum_attribute("orderState", "order_state"),
        "changePaymentState": update_enum_attribute("paymentState", "payment_state"),
        "changeShipmentState": update_enum_attribute("shipmentState", "shipment_state"),
        "addDelivery": add_delivery(),
        "setShippingAddress": update_attribute("shippingAddress", "address"),
        "setBillingAddress": update_attribute("billingAddress", "address"),
        "setCustomerEmail": update_attribute("customerEmail", "email"),
        "setCustomField": set_custom_field(),
        "addPayment": add_payment(),
    }



# DO NOT EDIT! This file is automatically generated
import datetime
import enum
import typing

from commercetools.types._abstract import _BaseType
from commercetools.types._common import (
    BaseResource,
    Reference,
    ReferenceTypeId,
    ResourceIdentifier,
)

if typing.TYPE_CHECKING:
    from ._common import CreatedBy, LastModifiedBy, LocalizedString
__all__ = [
    "State",
    "StateAddRolesAction",
    "StateChangeInitialAction",
    "StateChangeKeyAction",
    "StateChangeTypeAction",
    "StateDraft",
    "StatePagedQueryResponse",
    "StateReference",
    "StateRemoveRolesAction",
    "StateResourceIdentifier",
    "StateRoleEnum",
    "StateSetDescriptionAction",
    "StateSetNameAction",
    "StateSetRolesAction",
    "StateSetTransitionsAction",
    "StateTypeEnum",
    "StateUpdate",
    "StateUpdateAction",
]


class State(BaseResource):
    #: :class:`str`
    id: str
    #: :class:`int`
    version: int
    #: :class:`datetime.datetime` `(Named` ``createdAt`` `in Commercetools)`
    created_at: datetime.datetime
    #: :class:`datetime.datetime` `(Named` ``lastModifiedAt`` `in Commercetools)`
    last_modified_at: datetime.datetime
    #: Optional :class:`commercetools.types.LastModifiedBy` `(Named` ``lastModifiedBy`` `in Commercetools)`
    last_modified_by: typing.Optional["LastModifiedBy"]
    #: Optional :class:`commercetools.types.CreatedBy` `(Named` ``createdBy`` `in Commercetools)`
    created_by: typing.Optional["CreatedBy"]
    #: :class:`str`
    key: str
    #: :class:`commercetools.types.StateTypeEnum`
    type: "StateTypeEnum"
    #: Optional :class:`commercetools.types.LocalizedString`
    name: typing.Optional["LocalizedString"]
    #: Optional :class:`commercetools.types.LocalizedString`
    description: typing.Optional["LocalizedString"]
    #: :class:`bool`
    initial: bool
    #: :class:`bool` `(Named` ``builtIn`` `in Commercetools)`
    built_in: bool
    #: Optional list of :class:`commercetools.types.StateRoleEnum`
    roles: typing.Optional[typing.List["StateRoleEnum"]]
    #: Optional list of :class:`commercetools.types.StateReference`
    transitions: typing.Optional[typing.List["StateReference"]]

    def __init__(
        self,
        *,
        id: str,
        version: int,
        created_at: datetime.datetime,
        last_modified_at: datetime.datetime,
        key: str,
        type: "StateTypeEnum",
        initial: bool,
        built_in: bool,
        last_modified_by: typing.Optional["LastModifiedBy"] = None,
        created_by: typing.Optional["CreatedBy"] = None,
        name: typing.Optional["LocalizedString"] = None,
        description: typing.Optional["LocalizedString"] = None,
        roles: typing.Optional[typing.List["StateRoleEnum"]] = None,
        transitions: typing.Optional[typing.List["StateReference"]] = None
    ) -> None:
        self.id = id
        self.version = version
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        self.last_modified_by = last_modified_by
        self.created_by = created_by
        self.key = key
        self.type = type
        self.name = name
        self.description = description
        self.initial = initial
        self.built_in = built_in
        self.roles = roles
        self.transitions = transitions
        super().__init__(
            id=id,
            version=version,
            created_at=created_at,
            last_modified_at=last_modified_at,
        )

    def __repr__(self) -> str:
        return (
            "State(id=%r, version=%r, created_at=%r, last_modified_at=%r, last_modified_by=%r, created_by=%r, key=%r, type=%r, name=%r, description=%r, initial=%r, built_in=%r, roles=%r, transitions=%r)"
            % (
                self.id,
                self.version,
                self.created_at,
                self.last_modified_at,
                self.last_modified_by,
                self.created_by,
                self.key,
                self.type,
                self.name,
                self.description,
                self.initial,
                self.built_in,
                self.roles,
                self.transitions,
            )
        )


class StateDraft(_BaseType):
    #: :class:`str`
    key: str
    #: :class:`commercetools.types.StateTypeEnum`
    type: "StateTypeEnum"
    #: Optional :class:`commercetools.types.LocalizedString`
    name: typing.Optional["LocalizedString"]
    #: Optional :class:`commercetools.types.LocalizedString`
    description: typing.Optional["LocalizedString"]
    #: Optional :class:`bool`
    initial: typing.Optional[bool]
    #: Optional list of :class:`commercetools.types.StateRoleEnum`
    roles: typing.Optional[typing.List["StateRoleEnum"]]
    #: Optional list of :class:`commercetools.types.StateResourceIdentifier`
    transitions: typing.Optional[typing.List["StateResourceIdentifier"]]

    def __init__(
        self,
        *,
        key: str,
        type: "StateTypeEnum",
        name: typing.Optional["LocalizedString"] = None,
        description: typing.Optional["LocalizedString"] = None,
        initial: typing.Optional[bool] = None,
        roles: typing.Optional[typing.List["StateRoleEnum"]] = None,
        transitions: typing.Optional[typing.List["StateResourceIdentifier"]] = None
    ) -> None:
        self.key = key
        self.type = type
        self.name = name
        self.description = description
        self.initial = initial
        self.roles = roles
        self.transitions = transitions
        super().__init__()

    def __repr__(self) -> str:
        return (
            "StateDraft(key=%r, type=%r, name=%r, description=%r, initial=%r, roles=%r, transitions=%r)"
            % (
                self.key,
                self.type,
                self.name,
                self.description,
                self.initial,
                self.roles,
                self.transitions,
            )
        )


class StatePagedQueryResponse(_BaseType):
    #: :class:`int`
    limit: int
    #: :class:`int`
    count: int
    #: Optional :class:`int`
    total: typing.Optional[int]
    #: :class:`int`
    offset: int
    #: List of :class:`commercetools.types.State`
    results: typing.Sequence["State"]

    def __init__(
        self,
        *,
        limit: int,
        count: int,
        offset: int,
        results: typing.Sequence["State"],
        total: typing.Optional[int] = None
    ) -> None:
        self.limit = limit
        self.count = count
        self.total = total
        self.offset = offset
        self.results = results
        super().__init__()

    def __repr__(self) -> str:
        return (
            "StatePagedQueryResponse(limit=%r, count=%r, total=%r, offset=%r, results=%r)"
            % (self.limit, self.count, self.total, self.offset, self.results)
        )


class StateReference(Reference):
    #: Optional :class:`commercetools.types.State`
    obj: typing.Optional["State"]

    def __init__(self, *, id: str, obj: typing.Optional["State"] = None) -> None:
        self.obj = obj
        super().__init__(type_id=ReferenceTypeId.STATE, id=id)

    def __repr__(self) -> str:
        return "StateReference(type_id=%r, id=%r, obj=%r)" % (
            self.type_id,
            self.id,
            self.obj,
        )


class StateResourceIdentifier(ResourceIdentifier):
    def __init__(
        self, *, id: typing.Optional[str] = None, key: typing.Optional[str] = None
    ) -> None:
        super().__init__(type_id=ReferenceTypeId.STATE, id=id, key=key)

    def __repr__(self) -> str:
        return "StateResourceIdentifier(type_id=%r, id=%r, key=%r)" % (
            self.type_id,
            self.id,
            self.key,
        )


class StateRoleEnum(enum.Enum):
    REVIEW_INCLUDED_IN_STATISTICS = "ReviewIncludedInStatistics"
    RETURN_ = "Return"


class StateTypeEnum(enum.Enum):
    ORDER_STATE = "OrderState"
    LINE_ITEM_STATE = "LineItemState"
    PRODUCT_STATE = "ProductState"
    REVIEW_STATE = "ReviewState"
    PAYMENT_STATE = "PaymentState"


class StateUpdate(_BaseType):
    #: :class:`int`
    version: int
    #: :class:`list`
    actions: list

    def __init__(self, *, version: int, actions: list) -> None:
        self.version = version
        self.actions = actions
        super().__init__()

    def __repr__(self) -> str:
        return "StateUpdate(version=%r, actions=%r)" % (self.version, self.actions)


class StateUpdateAction(_BaseType):
    #: :class:`str`
    action: str

    def __init__(self, *, action: str) -> None:
        self.action = action
        super().__init__()

    def __repr__(self) -> str:
        return "StateUpdateAction(action=%r)" % (self.action,)


class StateAddRolesAction(StateUpdateAction):
    #: List of :class:`commercetools.types.StateRoleEnum`
    roles: typing.List["StateRoleEnum"]

    def __init__(self, *, roles: typing.List["StateRoleEnum"]) -> None:
        self.roles = roles
        super().__init__(action="addRoles")

    def __repr__(self) -> str:
        return "StateAddRolesAction(action=%r, roles=%r)" % (self.action, self.roles)


class StateChangeInitialAction(StateUpdateAction):
    #: :class:`bool`
    initial: bool

    def __init__(self, *, initial: bool) -> None:
        self.initial = initial
        super().__init__(action="changeInitial")

    def __repr__(self) -> str:
        return "StateChangeInitialAction(action=%r, initial=%r)" % (
            self.action,
            self.initial,
        )


class StateChangeKeyAction(StateUpdateAction):
    #: :class:`str`
    key: str

    def __init__(self, *, key: str) -> None:
        self.key = key
        super().__init__(action="changeKey")

    def __repr__(self) -> str:
        return "StateChangeKeyAction(action=%r, key=%r)" % (self.action, self.key)


class StateChangeTypeAction(StateUpdateAction):
    #: :class:`commercetools.types.StateTypeEnum`
    type: "StateTypeEnum"

    def __init__(self, *, type: "StateTypeEnum") -> None:
        self.type = type
        super().__init__(action="changeType")

    def __repr__(self) -> str:
        return "StateChangeTypeAction(action=%r, type=%r)" % (self.action, self.type)


class StateRemoveRolesAction(StateUpdateAction):
    #: List of :class:`commercetools.types.StateRoleEnum`
    roles: typing.List["StateRoleEnum"]

    def __init__(self, *, roles: typing.List["StateRoleEnum"]) -> None:
        self.roles = roles
        super().__init__(action="removeRoles")

    def __repr__(self) -> str:
        return "StateRemoveRolesAction(action=%r, roles=%r)" % (self.action, self.roles)


class StateSetDescriptionAction(StateUpdateAction):
    #: :class:`commercetools.types.LocalizedString`
    description: "LocalizedString"

    def __init__(self, *, description: "LocalizedString") -> None:
        self.description = description
        super().__init__(action="setDescription")

    def __repr__(self) -> str:
        return "StateSetDescriptionAction(action=%r, description=%r)" % (
            self.action,
            self.description,
        )


class StateSetNameAction(StateUpdateAction):
    #: :class:`commercetools.types.LocalizedString`
    name: "LocalizedString"

    def __init__(self, *, name: "LocalizedString") -> None:
        self.name = name
        super().__init__(action="setName")

    def __repr__(self) -> str:
        return "StateSetNameAction(action=%r, name=%r)" % (self.action, self.name)


class StateSetRolesAction(StateUpdateAction):
    #: List of :class:`commercetools.types.StateRoleEnum`
    roles: typing.List["StateRoleEnum"]

    def __init__(self, *, roles: typing.List["StateRoleEnum"]) -> None:
        self.roles = roles
        super().__init__(action="setRoles")

    def __repr__(self) -> str:
        return "StateSetRolesAction(action=%r, roles=%r)" % (self.action, self.roles)


class StateSetTransitionsAction(StateUpdateAction):
    #: Optional list of :class:`commercetools.types.StateResourceIdentifier`
    transitions: typing.Optional[typing.List["StateResourceIdentifier"]]

    def __init__(
        self,
        *,
        transitions: typing.Optional[typing.List["StateResourceIdentifier"]] = None
    ) -> None:
        self.transitions = transitions
        super().__init__(action="setTransitions")

    def __repr__(self) -> str:
        return "StateSetTransitionsAction(action=%r, transitions=%r)" % (
            self.action,
            self.transitions,
        )



import os
import typing
import urllib.parse

import requests
from marshmallow.base import SchemaABC
from oauthlib.oauth2 import BackendApplicationClient
from requests.adapters import HTTPAdapter
from requests_oauthlib import OAuth2Session
from urllib3.util.retry import Retry

from commercetools._schemas._error import ErrorResponseSchema
from commercetools.constants import HEADER_CORRELATION_ID
from commercetools.exceptions import CommercetoolsError
from commercetools.helpers import _concurrent_retry
from commercetools.services import ServicesMixin
from commercetools.utils import BaseTokenSaver, DefaultTokenSaver, fix_token_url


class RefreshingOAuth2Session(OAuth2Session):
    def refresh_token(self, token_url, **kwargs):
        kwargs.update(self.auto_refresh_kwargs)
        kwargs["scope"] = self.scope
        return self.fetch_token(token_url, **kwargs)


class Client(ServicesMixin):
    """The Commercetools Client, used to interact with the Commercetools API.

    :param project_key: the key for the project with which you want to interact
    :param client_id: the oauth2 client id
    :param client_secret: the oauth2 client secret
    :param scope: the oauth2 scope. If None then 'manage_project:{project_key}'
    :param url: the api endpoint
    :param token_url: the oauth2 token url endpoint. This should be the full
     path to the token url.
    :param token_saver: optional custom token saver to store and retrieve the
     oauth2 tokens.

    """

    def __init__(
        self,
        project_key: str = None,
        client_id: str = None,
        client_secret: str = None,
        scope: typing.List[str] = None,
        url: str = None,
        token_url: str = None,
        token_saver: BaseTokenSaver = None,
    ) -> None:

        # Use environment variables as fallback
        config = {
            "project_key": project_key,
            "client_id": client_id,
            "client_secret": client_secret,
            "url": url,
            "token_url": token_url,
            "scope": scope,
        }
        # Make sure we use the config vars
        del project_key, client_id, client_secret, url, token_url, scope

        self._config = self._read_env_vars(config)
        self._config["token_url"] = fix_token_url(self._config["token_url"])
        self._token_saver = token_saver or DefaultTokenSaver()
        self._url = self._config["url"]
        self._base_url = f"{self._config['url']}/{self._config['project_key']}/"

        # Fetch token from the token saver
        token = self._token_saver.get_token(
            self._config["client_id"], self._config["scope"]
        )

        client = BackendApplicationClient(
            client_id=self._config["client_id"], scope=self._config["scope"]
        )
        self._http_client = RefreshingOAuth2Session(
            client=client,
            scope=self._config["scope"],
            auto_refresh_url=self._config["token_url"],
            auto_refresh_kwargs={
                "client_id": self._config["client_id"],
                "client_secret": self._config["client_secret"],
            },
            token_updater=self._save_token,
        )

        # Register retry handling for Connection errors and 502, 503, 504.
        retry = Retry(status=3, connect=3, status_forcelist=[502, 503, 504])
        adapter = HTTPAdapter(max_retries=retry)
        self._http_client.mount("http://", adapter)
        self._http_client.mount("https://", adapter)

        if token:
            self._http_client.token = token
        else:
            token = self._http_client.fetch_token(
                token_url=self._config["token_url"],
                scope=self._config["scope"],
                client_id=self._config["client_id"],
                client_secret=self._config["client_secret"],
            )
            self._save_token(token)

    def _save_token(self, token):
        self._token_saver.add_token(
            self._config["client_id"], self._config["scope"], token
        )

    def _get(
        self,
        endpoint: str,
        params: typing.Dict[str, typing.Any],
        schema_cls: typing.Type[SchemaABC],
    ) -> typing.Any:
        """Retrieve a single object from the commercetools platform"""
        response = self._http_client.get(self._base_url + endpoint, params=params)

        if response.status_code == 200:
            return schema_cls().load(response.json())
        return self._process_error(response)

    def _post(
        self,
        endpoint: str,
        params: typing.Dict[str, str],
        data_object: typing.Any,
        request_schema_cls: typing.Optional[typing.Type[SchemaABC]],
        response_schema_cls: typing.Type[SchemaABC],
        form_encoded: bool = False,
        force_update: bool = False,
    ) -> typing.Any:
        """Retrieve a single object from the commercetools platform"""

        @_concurrent_retry(3 if force_update else 0)
        def remote_http_call(data):
            if form_encoded and data is not None:
                kwargs = {"data": data}
            else:
                kwargs = {"json": data}
            if params:
                kwargs["params"] = params

            response = self._http_client.post(self._base_url + endpoint, **kwargs)
            if response.status_code in (200, 201):
                return response_schema_cls().load(response.json())
            return self._process_error(response)

        if request_schema_cls is not None:
            data = request_schema_cls().dump(data_object)
        else:
            data = None
        return remote_http_call(data)

    def _upload(
        self,
        endpoint: str,
        params: typing.Dict[str, str],
        file: typing.IO,
        response_schema_cls: typing.Type[SchemaABC],
    ) -> typing.Any:
        """Retrieve a single object from the commercetools platform"""
        response = self._http_client.post(
            self._base_url + endpoint, data=file.read(), params=params
        )

        if response.status_code in (200, 201):
            return response_schema_cls().load(response.json())
        return self._process_error(response)

    def _delete(
        self,
        endpoint: str,
        params: typing.Dict[str, str],
        response_schema_cls: typing.Type[SchemaABC],
        force_delete: bool = False,
    ) -> typing.Any:
        """Delete an object from the commercetools platform"""

        @_concurrent_retry(3 if force_delete else 0)
        def remote_http_call(data):
            response = self._http_client.delete(
                self._base_url + endpoint, params=params
            )
            if response.status_code == 200:
                return response_schema_cls().load(response.json())
            return self._process_error(response)

        return remote_http_call(params)

    def _process_error(self, response: requests.Response) -> None:
        correlation_id = response.headers.get(HEADER_CORRELATION_ID)
        if not response.content:
            response.raise_for_status()
        obj = ErrorResponseSchema().loads(response.content)

        # We'll fetch the 'raw' errors from the response because some of the
        # attributes are not included in the schemas.
        # With the raw errors in the CommercetoolsError object we can use that
        # information later to render more detailed error messages
        errors_raw = []
        try:
            response_json = response.json()
        except ValueError:
            pass
        else:
            errors_raw = response_json.get("errors", [])

        raise CommercetoolsError(obj.message, errors_raw, obj, correlation_id)

    def _read_env_vars(self, config: dict) -> dict:
        if not config.get("project_key"):
            config["project_key"] = os.environ.get("CTP_PROJECT_KEY")

        if not config.get("client_id"):
            config["client_id"] = os.environ.get("CTP_CLIENT_ID")

        if not config.get("client_secret"):
            config["client_secret"] = os.environ.get("CTP_CLIENT_SECRET")

        if not config.get("url"):
            config["url"] = os.environ.get("CTP_API_URL")

        if not config.get("token_url"):
            config["token_url"] = os.environ.get("CTP_AUTH_URL")

        if not config["scope"]:
            config["scope"] = os.environ.get("CTP_SCOPES")
            if config["scope"]:
                config["scope"] = config["scope"].split(",")
            else:
                config["scope"] = ["manage_project:%s" % config["project_key"]]

        for key, value in config.items():
            if value is None:
                raise ValueError(f"No value set for {key}")

        return config



# DO NOT EDIT! This file is automatically generated
import datetime
import typing

from commercetools.types._abstract import _BaseType
from commercetools.types._common import BaseResource, Reference, ReferenceTypeId

if typing.TYPE_CHECKING:
    from ._common import CreatedBy, LastModifiedBy
__all__ = [
    "CustomObject",
    "CustomObjectDraft",
    "CustomObjectPagedQueryResponse",
    "CustomObjectReference",
]


class CustomObject(BaseResource):
    #: :class:`str`
    id: str
    #: :class:`int`
    version: int
    #: :class:`datetime.datetime` `(Named` ``createdAt`` `in Commercetools)`
    created_at: datetime.datetime
    #: :class:`datetime.datetime` `(Named` ``lastModifiedAt`` `in Commercetools)`
    last_modified_at: datetime.datetime
    #: Optional :class:`commercetools.types.LastModifiedBy` `(Named` ``lastModifiedBy`` `in Commercetools)`
    last_modified_by: typing.Optional["LastModifiedBy"]
    #: Optional :class:`commercetools.types.CreatedBy` `(Named` ``createdBy`` `in Commercetools)`
    created_by: typing.Optional["CreatedBy"]
    #: :class:`str`
    container: str
    #: :class:`str`
    key: str
    #: :class:`typing.Any`
    value: typing.Any

    def __init__(
        self,
        *,
        id: str,
        version: int,
        created_at: datetime.datetime,
        last_modified_at: datetime.datetime,
        container: str,
        key: str,
        value: typing.Any,
        last_modified_by: typing.Optional["LastModifiedBy"] = None,
        created_by: typing.Optional["CreatedBy"] = None
    ) -> None:
        self.id = id
        self.version = version
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        self.last_modified_by = last_modified_by
        self.created_by = created_by
        self.container = container
        self.key = key
        self.value = value
        super().__init__(
            id=id,
            version=version,
            created_at=created_at,
            last_modified_at=last_modified_at,
        )

    def __repr__(self) -> str:
        return (
            "CustomObject(id=%r, version=%r, created_at=%r, last_modified_at=%r, last_modified_by=%r, created_by=%r, container=%r, key=%r, value=%r)"
            % (
                self.id,
                self.version,
                self.created_at,
                self.last_modified_at,
                self.last_modified_by,
                self.created_by,
                self.container,
                self.key,
                self.value,
            )
        )


class CustomObjectDraft(_BaseType):
    #: :class:`str`
    container: str
    #: :class:`str`
    key: str
    #: :class:`typing.Any`
    value: typing.Any
    #: Optional :class:`int`
    version: typing.Optional[int]

    def __init__(
        self,
        *,
        container: str,
        key: str,
        value: typing.Any,
        version: typing.Optional[int] = None
    ) -> None:
        self.container = container
        self.key = key
        self.value = value
        self.version = version
        super().__init__()

    def __repr__(self) -> str:
        return "CustomObjectDraft(container=%r, key=%r, value=%r, version=%r)" % (
            self.container,
            self.key,
            self.value,
            self.version,
        )


class CustomObjectPagedQueryResponse(_BaseType):
    #: :class:`int`
    limit: int
    #: :class:`int`
    count: int
    #: Optional :class:`int`
    total: typing.Optional[int]
    #: :class:`int`
    offset: int
    #: List of :class:`commercetools.types.CustomObject`
    results: typing.Sequence["CustomObject"]

    def __init__(
        self,
        *,
        limit: int,
        count: int,
        offset: int,
        results: typing.Sequence["CustomObject"],
        total: typing.Optional[int] = None
    ) -> None:
        self.limit = limit
        self.count = count
        self.total = total
        self.offset = offset
        self.results = results
        super().__init__()

    def __repr__(self) -> str:
        return (
            "CustomObjectPagedQueryResponse(limit=%r, count=%r, total=%r, offset=%r, results=%r)"
            % (self.limit, self.count, self.total, self.offset, self.results)
        )


class CustomObjectReference(Reference):
    #: Optional :class:`commercetools.types.CustomObject`
    obj: typing.Optional["CustomObject"]

    def __init__(self, *, id: str, obj: typing.Optional["CustomObject"] = None) -> None:
        self.obj = obj
        super().__init__(type_id=ReferenceTypeId.KEY_VALUE_DOCUMENT, id=id)

    def __repr__(self) -> str:
        return "CustomObjectReference(type_id=%r, id=%r, obj=%r)" % (
            self.type_id,
            self.id,
            self.obj,
        )
